# ***************************************************************************
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM calculix constraint sectionprint"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"


def get_analysis_types():
    return "all"  # write for all analysis types


def get_sets_name():
    return "constraints_sectionprint_surface_sets"


def get_constraint_title():
    return "SectionPrint Constraints"


def get_before_write_meshdata_constraint():
    return ""


def get_after_write_meshdata_constraint():
    return ""


def get_before_write_constraint():
    return ""


def get_after_write_constraint():
    return ""


def write_meshdata_constraint(f, femobj, sectionprint_obj, ccxwriter):
    f.write(f"*SURFACE, NAME=SECTIONFACE{sectionprint_obj.Name}\n")
    for i in femobj["SectionPrintFaces"]:
        f.write(f"{i[0]},S{i[1]}\n")


def write_constraint(f, femobj, sectionprint_obj, ccxwriter):

    # floats read from ccx should use {:.13G}, see comment in writer module

    variable = sectionprint_obj.Variable
    if variable == "Section Force":
        key = "SOF, SOM, SOAREA"
    elif variable == "Heat Flux":
        key = "FLUX"
    elif variable == "Drag Stress":
        key = "DRAG"

    f.write(
        "*SECTION PRINT, SURFACE=SECTIONFACE{}, NAME=SECTIONPRINT{}\n".format(
            sectionprint_obj.Name, sectionprint_obj.Name
        )
    )
    f.write(key + "\n")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create Clone objects."""
## @package make_clone
# \ingroup draftmake
# \brief Provides functions to create Clone objects.

## \addtogroup draftmake
# @{
import FreeCAD as App
from draftobjects.clone import Clone
from draftutils import params
from draftutils import utils
from draftutils import gui_utils

if App.GuiUp:
    from draftviewproviders.view_clone import ViewProviderClone


def make_clone(obj, delta=None, forcedraft=False):
    """clone(obj,[delta,forcedraft])

    Makes a clone of the given object(s).
    The clone is an exact, linked copy of the given object. If the original
    object changes, the final object changes too.

    Parameters
    ----------
    obj :

    delta : Base.Vector
        Delta Vector to move the clone from the original position.

    forcedraft : bool
        If forcedraft is True, the resulting object is a Draft clone
        even if the input object is an Arch object.

    """

    prefix = params.get_param("ClonePrefix")

    cl = None

    if prefix:
        prefix = prefix.strip() + " "

    if not isinstance(obj, list):
        obj = [obj]

    if (
        len(obj) == 1
        and obj[0].isDerivedFrom("Part::Part2DObject")
        and utils.get_type(obj[0]) not in ["BezCurve", "BSpline", "Wire"]
    ):
        # "BezCurve", "BSpline" and "Wire" objects created with < v1.1
        # are "Part::Part2DObject" objects but they need not be 2D.
        cl = App.ActiveDocument.addObject("Part::Part2DObjectPython", "Clone2D")
        cl.Label = prefix + obj[0].Label + " (2D)"
    elif (
        len(obj) == 1
        and (hasattr(obj[0], "CloneOf") or utils.get_type(obj[0]) == "BuildingPart")
        and not forcedraft
    ):
        # arch objects can be clones
        try:
            import Arch
        except:
            # BIM not present
            pass
        else:
            if utils.get_type(obj[0]) == "BuildingPart":
                cl = Arch.makeComponent()
            else:
                try:  # new-style make function
                    cl = getattr(Arch, "make_" + obj[0].Proxy.Type.lower())()
                except Exception:
                    try:  # old-style make function
                        cl = getattr(Arch, "make" + obj[0].Proxy.Type)()
                    except Exception:
                        pass  # not a standard Arch object... Fall back to Draft mode
            if cl:
                base = utils.get_clone_base(obj[0])
                cl.Label = prefix + base.Label
                cl.CloneOf = base
                if utils.get_type(obj[0]) != "BuildingPart":
                    cl.Placement = obj[0].Placement
                for prop in ("Description", "IfcType", "Material", "Subvolume", "Tag"):
                    try:
                        setattr(cl, prop, getattr(base, prop))
                    except Exception:
                        pass
                if App.GuiUp:
                    gui_utils.format_object(cl, base)
                    gui_utils.select(cl)
                return cl

    # fall back to Draft clone mode
    if not cl:
        cl = App.ActiveDocument.addObject("Part::FeaturePython", "Clone")
        cl.addExtension("Part::AttachExtensionPython")
        cl.Label = prefix + obj[0].Label
    Clone(cl)
    cl.Objects = obj
    if delta:
        cl.Placement.move(delta)
    elif (len(obj) == 1) and hasattr(obj[0], "Placement"):
        cl.Placement = obj[0].Placement
    if hasattr(cl, "LongName") and hasattr(obj[0], "LongName"):
        cl.LongName = obj[0].LongName
    if App.GuiUp:
        ViewProviderClone(cl.ViewObject)
        gui_utils.format_object(cl, obj[0])
        gui_utils.select(cl)
    return cl


clone = make_clone

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2022 Larry Woestman <LarryWoestman2@gmail.com>          *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""
These are common functions and classes for creating custom post processors.
"""


from Path.Base.MachineState import MachineState
from PySide import QtCore, QtGui
import FreeCAD
import Part
import Path
import os
import re

debug = False
if debug:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate

FreeCADGui = None
if FreeCAD.GuiUp:
    import FreeCADGui


class FilenameGenerator:
    def __init__(self, job):
        self.job = job
        self.subpartname = ""
        self.sequencenumber = 0
        path, filename, ext = self.get_path_and_filename_default()

        self.qualified_path = self._apply_path_substitutions(path)
        self.qualified_filename = self._apply_filename_substitutions(filename)
        self.extension = ext

    def get_path_and_filename_default(self):
        outputpath = ""
        filename = ""
        ext = ".nc"

        validPathSubstitutions = ["D", "d", "M", "j"]
        validFilenameSubstitutions = ["j", "d", "T", "t", "W", "O", "S"]

        if self.job.PostProcessorOutputFile:
            candidateOutputPath, candidateFilename = os.path.split(self.job.PostProcessorOutputFile)

            if candidateOutputPath:
                outputpath = candidateOutputPath

            if candidateFilename:
                filename, ext = os.path.splitext(candidateFilename)
        else:
            outputpath, filename = os.path.split(Path.Preferences.defaultOutputFile())
            filename, ext = os.path.splitext(filename)

        # Make sure we have something to work with
        if not filename:
            filename = FreeCAD.ActiveDocument.Label

        if not outputpath:
            outputpath, _ = os.path.split(FreeCAD.ActiveDocument.getFileName())

        if not outputpath:
            outputpath = (
                os.getcwd()
            )  ## TODO: This should be avoided as it gives the Freecad executable's path in some systems (e.g. Windows)

        if not ext:
            ext = ".nc"

        # Check for invalid matches
        for match in re.findall(r"%(.)", outputpath):
            Path.Log.debug(f"match: {match}")
            if match not in validPathSubstitutions:
                outputpath = outputpath.replace(f"%{match}", "")
                FreeCAD.Console.PrintWarning(
                    "Invalid substitution strings will be ignored in output path: %s\n" % match
                )

        for match in re.findall(r"%(.)", filename):
            Path.Log.debug(f"match: {match}")
            if match not in validFilenameSubstitutions:
                filename = filename.replace(f"%{match}", "")
                FreeCAD.Console.PrintWarning(
                    "Invalid substitution strings will be ignored in file path: %s\n" % match
                )

        Path.Log.debug(f"outputpath: {outputpath} filename: {filename} ext: {ext}")
        return outputpath, filename, ext

    def set_subpartname(self, subpartname):
        self.subpartname = subpartname

    def _apply_path_substitutions(self, file_path):
        """Apply substitutions based on job settings and other parameters."""
        substitutions = {
            "%D": os.path.dirname(self.job.Document.FileName or "."),
            "%d": self.job.Document.Label,
            "%j": self.job.Label,
            "%M": os.path.dirname(FreeCAD.getUserMacroDir()),
        }
        for key, value in substitutions.items():
            file_path = file_path.replace(key, value)

        Path.Log.debug(f"file_path: {file_path}")
        return file_path

    def _apply_filename_substitutions(self, file_name):
        Path.Log.debug(f"file_name: {file_name}")
        """Apply substitutions based on job settings and other parameters."""
        substitutions = {
            "%d": self.job.Document.Label,
            "%j": self.job.Label,
            "%T": self.subpartname,  # Tool Number
            "%t": self.subpartname,  # Tool Controller Label
            "%W": self.subpartname,  # Fixture
            "%O": self.subpartname,  # Operation
        }
        for key, value in substitutions.items():
            file_name = file_name.replace(key, value)

        Path.Log.debug(f"file_name: {file_name}")
        return file_name

    def generate_filenames(self):
        """Yield filenames indefinitely with proper substitutions."""
        while True:
            temp_filename = self.qualified_filename
            Path.Log.debug(f"temp_filename: {temp_filename}")
            explicit_sequence = False
            matches = re.findall(r"%S", temp_filename)
            if matches:
                Path.Log.debug(f"matches: {matches}")
                temp_filename = re.sub(r"%S", str(self.sequencenumber), temp_filename)
                explicit_sequence = True

            subpart = f"-{self.subpartname}" if self.subpartname else ""
            sequence = (
                f"-{self.sequencenumber}" if not explicit_sequence and self.sequencenumber else ""
            )
            filename = f"{temp_filename}{subpart}{sequence}{self.extension}"

            # Trim leading dash if filename starts with one
            if filename.startswith("-"):
                filename = filename[1:]
            full_path = os.path.join(self.qualified_path, filename)

            self.sequencenumber += 1
            Path.Log.debug(f"yielding filename: {full_path}")
            yield os.path.normpath(full_path)


class GCodeHighlighter(QtGui.QSyntaxHighlighter):
    def __init__(self, parent=None):
        super(GCodeHighlighter, self).__init__(parent)

        keywordFormat = QtGui.QTextCharFormat()
        keywordFormat.setForeground(QtCore.Qt.cyan)
        keywordFormat.setFontWeight(QtGui.QFont.Bold)
        keywordPatterns = ["\\bG[0-9]+\\b", "\\bM[0-9]+\\b"]

        self.highlightingRules = [
            (QtCore.QRegularExpression(pattern), keywordFormat) for pattern in keywordPatterns
        ]

        speedFormat = QtGui.QTextCharFormat()
        speedFormat.setFontWeight(QtGui.QFont.Bold)
        speedFormat.setForeground(QtCore.Qt.green)
        self.highlightingRules.append((QtCore.QRegularExpression("\\bF[0-9\\.]+\\b"), speedFormat))

    def highlightBlock(self, text):
        for pattern, hlFormat in self.highlightingRules:
            expression = QtCore.QRegularExpression(pattern)
            index = expression.match(text)
            while index.hasMatch():
                length = index.capturedLength()
                self.setFormat(index.capturedStart(), length, hlFormat)
                index = expression.match(text, index.capturedStart() + length)


class GCodeEditorDialog(QtGui.QDialog):
    def __init__(self, parent=None):
        if parent is None:
            parent = FreeCADGui.getMainWindow()
        QtGui.QDialog.__init__(self, parent)

        layout = QtGui.QVBoxLayout(self)

        # nice text editor widget for editing the gcode
        self.editor = QtGui.QTextEdit()
        font = QtGui.QFont()
        font.setFamily("Courier")
        font.setFixedPitch(True)
        font.setPointSize(10)
        self.editor.setFont(font)
        self.editor.setText("G01 X55 Y4.5 F300.0")
        layout.addWidget(self.editor)

        # OK and Cancel buttons
        self.buttons = QtGui.QDialogButtonBox(
            QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel,
            QtCore.Qt.Horizontal,
            self,
        )
        layout.addWidget(self.buttons)

        # restore placement and size
        self.paramKey = "User parameter:BaseApp/Values/Mod/CAM/GCodeEditor/"
        params = FreeCAD.ParamGet(self.paramKey)
        posX = params.GetInt("posX")
        posY = params.GetInt("posY")
        if posX > 0 and posY > 0:
            self.move(posX, posY)
        width = params.GetInt("width")
        height = params.GetInt("height")
        if width > 0 and height > 0:
            self.resize(width, height)

        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)

    def done(self, *args, **kwargs):
        params = FreeCAD.ParamGet(self.paramKey)
        params.SetInt("posX", self.x())
        params.SetInt("posY", self.y())
        params.SetInt("width", self.size().width())
        params.SetInt("height", self.size().height())
        return QtGui.QDialog.done(self, *args, **kwargs)


def stringsplit(commandline):
    returndict = {
        "command": None,
        "X": None,
        "Y": None,
        "Z": None,
        "A": None,
        "B": None,
        "F": None,
        "T": None,
        "S": None,
        "I": None,
        "J": None,
        "K": None,
        "txt": None,
    }
    wordlist = [a.strip() for a in commandline.split(" ")]
    if wordlist[0][0] == "(":
        returndict["command"] = "message"
        returndict["txt"] = wordlist[0]
    else:
        returndict["command"] = wordlist[0]
    for word in wordlist[1:]:
        returndict[word[0]] = word[1:]

    return returndict


def fmt(num, dec, units):
    """Use to format axis moves, feedrate, etc for decimal places and units."""
    if units == "G21":  # metric
        fnum = "%.*f" % (dec, num)
    else:  # inch
        fnum = "%.*f" % (dec, num / 25.4)  # since FreeCAD uses metric units internally
    return fnum


def editor(gcode):
    """Pops up a handy little editor to look at the code output."""
    prefs = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/CAM")
    # default Max Highlighter Size = 512 Ko
    defaultMHS = 512 * 1024
    mhs = prefs.GetUnsigned("inspecteditorMaxHighlighterSize", defaultMHS)

    dia = GCodeEditorDialog()
    dia.editor.setText(gcode)
    gcodeSize = len(dia.editor.toPlainText())
    if gcodeSize <= mhs:
        # because of poor performance, syntax highlighting is
        # limited to mhs octets (default 512 KB).
        # It seems than the response time curve has an inflexion near 500 KB
        # beyond 500 KB, the response time increases exponentially.
        dia.highlighter = GCodeHighlighter(dia.editor.document())
    else:
        FreeCAD.Console.PrintMessage(
            translate(
                "Path",
                "GCode size too big ({} o), disabling syntax highlighter.".format(gcodeSize),
            )
        )
    result = dia.exec_()
    if result:  # If user selected 'OK' get modified G Code
        final = dia.editor.toPlainText()
    else:
        final = gcode
    return final


def fcoms(string, commentsym):
    """Filter and rebuild comments with user preferred comment symbol."""
    if len(commentsym) == 1:
        s1 = string.replace("(", commentsym)
        comment = s1.replace(")", "")
    else:
        return string
    return comment


def splitArcs(path):
    """Filter a path object and replace all G2/G3 moves with discrete G1 moves.

    Returns a Path object.
    """
    prefGrp = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/CAM")
    deflection = prefGrp.GetFloat("LibAreaCurveAccuarcy", 0.01)

    results = []
    if not isinstance(path, Path.Path):
        raise TypeError("path must be a Path object")

    machine = MachineState()
    for command in path.Commands:

        if command.Name not in Path.Geom.CmdMoveArc:
            machine.addCommand(command)
            results.append(command)
            continue

        edge = Path.Geom.edgeForCmd(command, machine.getPosition())
        pts = edge.discretize(Deflection=deflection)
        edges = [Part.makeLine(v1, v2) for v1, v2 in zip(pts, pts[1:])]
        for edge in edges:
            results.extend(Path.Geom.cmdsForEdge(edge))

        machine.addCommand(command)

    return Path.Path(results)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""ToolBit selector dialog."""

from PySide import QtWidgets
import FreeCAD
from ...camassets import cam_assets
from ...toolbit import ToolBit
from .browser import ToolBitBrowserWidget


class ToolBitSelector(QtWidgets.QDialog):
    """
    A dialog for selecting ToolBits using the ToolBitBrowserWidget.
    """

    def __init__(
        self, parent=None, compact=False, button_label=FreeCAD.Qt.translate("CAM", "Add Tool")
    ):
        super().__init__(parent)

        self.setMinimumSize(600, 400)

        self.setWindowTitle(FreeCAD.Qt.translate("CAM", "Select Toolbit"))

        self._browser_widget = ToolBitBrowserWidget(cam_assets, compact=compact)

        # Create OK and Cancel buttons
        self._ok_button = QtWidgets.QPushButton(button_label)
        self._cancel_button = QtWidgets.QPushButton("Cancel")

        # Connect buttons to their actions
        self._ok_button.clicked.connect(self.accept)
        self._cancel_button.clicked.connect(self.reject)

        # Layout setup
        layout = QtWidgets.QVBoxLayout(self)
        layout.addWidget(self._browser_widget)

        button_layout = QtWidgets.QHBoxLayout()
        button_layout.addStretch()
        button_layout.addWidget(self._cancel_button)
        button_layout.addWidget(self._ok_button)

        layout.addLayout(button_layout)

        # Disable OK button initially until a tool is selected
        self._ok_button.setEnabled(False)
        self._browser_widget.toolSelected.connect(self._on_tool_selected)
        self._browser_widget.itemDoubleClicked.connect(self.accept)

        self._selected_tool_uri = None

    def _on_tool_selected(self, uri):
        """Enables/disables OK button based on selection."""
        self._selected_tool_uri = uri
        self._ok_button.setEnabled(uri is not None)

    def get_selected_tool_uri(self):
        """Returns the URI of the selected tool bit."""
        return self._selected_tool_uri

    def get_selected_tool(self) -> ToolBit:
        """Returns the selected ToolBit object, or None if none selected."""
        uri = self.get_selected_tool_uri()
        if uri:
            # Assuming ToolBit.from_uri exists and loads the ToolBit object
            return cam_assets.get(uri)
        return None



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to create orthogonal 2D and 3D Array objects."""
## @package make_orthoarray
# \ingroup draftmake
# \brief Provides functions to create orthogonal 2D and 3D Arrays.

## \addtogroup draftmake
# @{
import FreeCAD as App
import draftutils.utils as utils
import draftmake.make_array as make_array

from draftutils.messages import _wrn, _err
from draftutils.translate import translate


def _make_ortho_array(
    base_object,
    v_x=App.Vector(10, 0, 0),
    v_y=App.Vector(0, 10, 0),
    v_z=App.Vector(0, 0, 10),
    n_x=2,
    n_y=2,
    n_z=1,
    use_link=True,
):
    """Create an orthogonal array from the given object.

    This is a simple wrapper of the `draftmake.make_array.make_array`
    function to be used by the different orthogonal arrays.

    - `make_ortho_array`
    - `make_ortho_array2d`, no Z direction
    - `make_rect_array`, strictly rectangular
    - `make_rect_array2d`, strictly rectangular, no Z direction

    This function has no error checking, nor does it display messages.
    This should be handled by the subfunctions that use this one.
    """
    _name = "_make_ortho_array"

    new_obj = make_array.make_array(
        base_object, arg1=v_x, arg2=v_y, arg3=v_z, arg4=n_x, arg5=n_y, arg6=n_z, use_link=use_link
    )
    return new_obj


def _are_vectors(v_x, v_y, v_z=None, name="Unknown"):
    """Check that the vectors are numbers."""
    try:
        if v_z:
            utils.type_check(
                [
                    (v_x, (int, float, App.Vector)),
                    (v_y, (int, float, App.Vector)),
                    (v_z, (int, float, App.Vector)),
                ],
                name=name,
            )
        else:
            utils.type_check(
                [(v_x, (int, float, App.Vector)), (v_y, (int, float, App.Vector))], name=name
            )
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number or vector."))
        return False, v_x, v_y, v_z

    if not isinstance(v_x, App.Vector):
        v_x = App.Vector(v_x, 0, 0)
        _wrn(translate("draft", "Input: single value expanded to vector."))
    if not isinstance(v_y, App.Vector):
        v_y = App.Vector(0, v_y, 0)
        _wrn(translate("draft", "Input: single value expanded to vector."))
    if v_z and not isinstance(v_z, App.Vector):
        v_z = App.Vector(0, 0, v_z)
        _wrn(translate("draft", "Input: single value expanded to vector."))

    return True, v_x, v_y, v_z


def _are_integers(n_x, n_y, n_z=None, name="Unknown"):
    """Check that the numbers are integers, with minimum value of 1."""
    try:
        if n_z:
            utils.type_check([(n_x, int), (n_y, int), (n_z, int)], name=name)
        else:
            utils.type_check([(n_x, int), (n_y, int)], name=name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be an integer number."))
        return False, n_x, n_y, n_z

    if n_x < 1:
        _wrn(translate("draft", "Input: number of elements must be at least 1. It is set to 1."))
        n_x = 1
    if n_y < 1:
        _wrn(translate("draft", "Input: number of elements must be at least 1. It is set to 1."))
        n_y = 1
    if n_z and n_z < 1:
        _wrn(translate("draft", "Input: number of elements must be at least 1. It is set to 1."))
        n_z = 1

    return True, n_x, n_y, n_z


def _are_numbers(d_x, d_y, d_z=None, name="Unknown"):
    """Check that the numbers are numbers."""
    try:
        if d_z:
            utils.type_check(
                [(d_x, (int, float)), (d_y, (int, float)), (d_z, (int, float))], name=name
            )
        else:
            utils.type_check([(d_x, (int, float)), (d_y, (int, float))], name=name)
    except TypeError:
        _err(translate("draft", "Wrong input: must be a number."))
        return False, d_x, d_y, d_z

    return True, d_x, d_y, d_z


def _find_object_in_doc(base_object, doc=None):
    """Check that a document is available and the object exists."""
    found, base_object = utils.find_object(base_object, doc=doc)
    if not found:
        _err(translate("draft", "Wrong input: base_object not in document."))

    return found, base_object


def make_ortho_array(
    base_object,
    v_x=App.Vector(10, 0, 0),
    v_y=App.Vector(0, 10, 0),
    v_z=App.Vector(0, 0, 10),
    n_x=2,
    n_y=2,
    n_z=1,
    use_link=True,
):
    """Create an orthogonal array from the given object.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    v_x, v_y, v_z: Base::Vector3, optional
        The vector indicating the vector displacement between two elements
        in the specified orthogonal direction X, Y, Z.

        By default:
        ::
            v_x = App.Vector(10, 0, 0)
            v_y = App.Vector(0, 10, 0)
            v_z = App.Vector(0, 0, 10)

        Given that this is a vectorial displacement
        the next object can appear displaced in one, two or three axes
        at the same time.

        For example
        ::
            v_x = App.Vector(10, 5, 0)

        means that the next element in the X direction will be displaced
        10 mm in X, 5 mm in Y, and 0 mm in Z.

        A traditional "rectangular" array is obtained when
        the displacement vector only has its corresponding component,
        like in the default case.

        If these values are entered as single numbers instead
        of vectors, the single value is expanded into a vector
        of the corresponding direction, and the other components are assumed
        to be zero.

        For example
        ::
            v_x = 15
            v_y = 10
            v_z = 1
        becomes
        ::
            v_x = App.Vector(15, 0, 0)
            v_y = App.Vector(0, 10, 0)
            v_z = App.Vector(0, 0, 1)

    n_x, n_y, n_z: int, optional
        The number of copies in the specified orthogonal direction X, Y, Z.
        This number includes the original object, therefore, it must be
        at least 1.

        The values of `n_x` and `n_y` default to 2,
        while `n_z` defaults to 1.
        This means the array is a planar array by default.

    use_link: bool, optional
        It defaults to `True`.
        If it is `True` the produced copies are not `Part::TopoShape` copies,
        but rather `App::Link` objects.
        The Links repeat the shape of the original `base_object` exactly,
        and therefore the resulting array is more memory efficient.

        Also, when `use_link` is `True`, the `Fuse` property
        of the resulting array does not work; the array doesn't
        contain separate shapes, it only has the original shape repeated
        many times, so there is nothing to fuse together.

        If `use_link` is `False` the original shape is copied many times.
        In this case the `Fuse` property is able to fuse
        all copies into a single object, if they touch each other.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'Array'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.

    See Also
    --------
    make_ortho_array2d, make_rect_array, make_rect_array2d, make_polar_array,
    make_circular_array, make_path_array, make_point_array
    """
    _name = "make_ortho_array"

    found, base_object = _find_object_in_doc(base_object, doc=App.activeDocument())
    if not found:
        return None

    ok, v_x, v_y, v_z = _are_vectors(v_x, v_y, v_z, name=_name)
    if not ok:
        return None

    ok, n_x, n_y, n_z = _are_integers(n_x, n_y, n_z, name=_name)
    if not ok:
        return None

    use_link = bool(use_link)

    new_obj = _make_ortho_array(
        base_object, v_x=v_x, v_y=v_y, v_z=v_z, n_x=n_x, n_y=n_y, n_z=n_z, use_link=use_link
    )
    return new_obj


def make_ortho_array2d(
    base_object, v_x=App.Vector(10, 0, 0), v_y=App.Vector(0, 10, 0), n_x=2, n_y=2, use_link=True
):
    """Create a 2D orthogonal array from the given object.

    This works the same as `make_ortho_array`.
    The Z component is ignored so it only considers vector displacements
    in X and Y directions.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    v_x, v_y: Base::Vector3, optional
        Vectorial displacement of elements
        in the corresponding X and Y directions.
        See `make_ortho_array`.

    n_x, n_y: int, optional
        Number of elements
        in the corresponding X and Y directions.
        See `make_ortho_array`.

    use_link: bool, optional
        If it is `True`, create `App::Link` array.
        See `make_ortho_array`.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'Array'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.

    See Also
    --------
    make_ortho_array, make_rect_array, make_rect_array2d, make_polar_array,
    make_circular_array, make_path_array, make_point_array
    """
    _name = "make_ortho_array2d"

    found, base_object = _find_object_in_doc(base_object, doc=App.activeDocument())
    if not found:
        return None

    ok, v_x, v_y, __ = _are_vectors(v_x, v_y, v_z=None, name=_name)
    if not ok:
        return None

    ok, n_x, n_y, __ = _are_integers(n_x, n_y, n_z=None, name=_name)
    if not ok:
        return None

    use_link = bool(use_link)

    new_obj = _make_ortho_array(base_object, v_x=v_x, v_y=v_y, n_x=n_x, n_y=n_y, use_link=use_link)
    return new_obj


def make_rect_array(base_object, d_x=10, d_y=10, d_z=10, n_x=2, n_y=2, n_z=1, use_link=True):
    """Create a rectangular array from the given object.

    This function wraps around `make_ortho_array`
    to produce strictly rectangular arrays, in which
    the displacement vectors `v_x`, `v_y`, and `v_z`
    only have their respective components in X, Y, and Z.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    d_x, d_y, d_z: Base::Vector3, optional
        Displacement of elements in the corresponding X, Y, and Z directions.

    n_x, n_y, n_z: int, optional
        Number of elements in the corresponding X, Y, and Z directions.

    use_link: bool, optional
        If it is `True`, create `App::Link` array.
        See `make_ortho_array`.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'Array'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.

    See Also
    --------
    make_ortho_array, make_ortho_array2d, make_rect_array2d, make_polar_array,
    make_circular_array, make_path_array, make_point_array
    """
    _name = "make_rect_array"

    found, base_object = _find_object_in_doc(base_object, doc=App.activeDocument())
    if not found:
        return None

    ok, d_x, d_y, d_z = _are_numbers(d_x, d_y, d_z, name=_name)
    if not ok:
        return None

    ok, n_x, n_y, n_z = _are_integers(n_x, n_y, n_z, _name)
    if not ok:
        return None

    use_link = bool(use_link)

    new_obj = _make_ortho_array(
        base_object,
        v_x=App.Vector(d_x, 0, 0),
        v_y=App.Vector(0, d_y, 0),
        v_z=App.Vector(0, 0, d_z),
        n_x=n_x,
        n_y=n_y,
        n_z=n_z,
        use_link=use_link,
    )
    return new_obj


def make_rect_array2d(base_object, d_x=10, d_y=10, n_x=2, n_y=2, use_link=True):
    """Create a 2D rectangular array from the given object.

    This function wraps around `make_ortho_array`,
    to produce strictly rectangular arrays, in which
    the displacement vectors `v_x` and `v_y`
    only have their respective components in X and Y.
    The Z component is ignored.

    Parameters
    ----------
    base_object: Part::Feature or str
        Any of object that has a `Part::TopoShape` that can be duplicated.
        This means most 2D and 3D objects produced with any workbench.
        If it is a string, it must be the `Label` of that object.
        Since a label is not guaranteed to be unique in a document,
        it will use the first object found with this label.

    d_x, d_y: Base::Vector3, optional
        Displacement of elements in the corresponding X and Y directions.

    n_x, n_y: int, optional
        Number of elements in the corresponding X and Y directions.

    use_link: bool, optional
        If it is `True`, create `App::Link` array.
        See `make_ortho_array`.

    Returns
    -------
    Part::FeaturePython
        A scripted object of type `'Array'`.
        Its `Shape` is a compound of the copies of the original object.

    None
        If there is a problem it will return `None`.

    See Also
    --------
    make_ortho_array, make_ortho_array2d, make_rect_array, make_polar_array,
    make_circular_array, make_path_array, make_point_array
    """
    _name = "make_rect_array2d"

    found, base_object = _find_object_in_doc(base_object, doc=App.activeDocument())
    if not found:
        return None

    ok, d_x, d_y, __ = _are_numbers(d_x, d_y, d_z=None, name=_name)
    if not ok:
        return None

    ok, n_x, n_y, __ = _are_integers(n_x, n_y, n_z=None, name=_name)
    if not ok:
        return None

    use_link = bool(use_link)

    new_obj = _make_ortho_array(
        base_object,
        v_x=App.Vector(d_x, 0, 0),
        v_y=App.Vector(0, d_y, 0),
        n_x=n_x,
        n_y=n_y,
        use_link=use_link,
    )
    return new_obj


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""Classes to manage the GUI presentation of installing an Addon (e.g., the sequence of dialog boxes
that do dependency resolution, error handling, etc.). See AddonInstallerGUI and MacroInstallerGUI
classes for details."""
import os
import sys
from typing import List

import addonmanager_freecad_interface as fci
from Widgets.addonmanager_utility_dialogs import MessageDialog
from addonmanager_toolbar_adapter import ToolbarAdapter

from PySideWrapper import QtCore, QtWidgets

from addonmanager_installer import AddonInstaller, MacroInstaller
from addonmanager_dependency_installer import DependencyInstaller
from addonmanager_metadata import Version
import addonmanager_utilities as utils
from Addon import Addon, MissingDependencies

translate = fci.translate

# pylint: disable=c-extension-no-member,too-few-public-methods,too-many-instance-attributes


class AddonInstallerGUI(QtCore.QObject):
    """GUI functions (sequence of dialog boxes) for installing an addon interactively. The actual
    installation is handled by the AddonInstaller class running in a separate QThread. An instance
    of this AddonInstallerGUI class should NOT be run in a separate thread but on the main GUI
    thread. All dialogs are modal."""

    # External classes are expected to "set and forget" this class, but in the event that some
    # action must be taken if the addon is actually installed, this signal is provided. Note that
    # this class already provides a "Successful installation" dialog, so external code need not
    # do so.
    success: QtCore.Signal = QtCore.Signal(object)

    # Emitted once all work has been completed, regardless of success or failure
    finished: QtCore.Signal = QtCore.Signal()

    def __init__(self, addon: Addon, addons: List[Addon] = None):
        super().__init__()
        self.addon_to_install = addon
        self.addons = [] if addons is None else addons
        self.installer = AddonInstaller(addon)
        self.dependency_installer = None
        self.install_worker = None
        self.dependency_dialog = None
        self.dependency_installation_dialog = None
        self.installing_dialog = None
        self.worker_thread = None

        # Set up the installer connections
        self.installer.success.connect(self._installation_succeeded)
        self.installer.failure.connect(self._installation_failed)

    def shutdown(self):
        try:
            self._stop_thread(self.worker_thread)
        except RuntimeError:
            # In some circumstances during shutdown the underlying C++ thread may already be gone:
            # we don't care
            pass

    @staticmethod
    def _stop_thread(thread: QtCore.QThread):
        if thread and hasattr(thread, "quit"):
            if thread.isRunning():
                fci.Console.PrintMessage(
                    "INTERNAL ERROR: a QThread is still running when it should have finished"
                )

                thread.requestInterruption()
                thread.wait(100)
            thread.quit()
            thread.wait(500)
            if thread.isRunning():
                fci.Console.PrintError(
                    "INTERNAL ERROR: Thread did not quit() cleanly, using terminate()\n"
                )
                thread.terminate()

    def run(self):
        """Instructs this class to begin displaying the necessary dialogs to guide a user through
        an Addon installation sequence. All dialogs are modal."""

        # Dependency check
        deps = MissingDependencies()
        deps.import_from_addon(self.addon_to_install, self.addons)

        self.dependency_installer = AddonDependencyInstallerGUI([self.addon_to_install], deps)
        self.dependency_installer.cancel.connect(self.finished.emit)
        self.dependency_installer.proceed.connect(self.install)
        self.dependency_installer.run()

    def install(self) -> None:
        """Installs or updates a workbench, macro, or package"""
        self.worker_thread = QtCore.QThread()
        self.worker_thread.setObjectName("Addon Installer worker thread")
        self.installer.moveToThread(self.worker_thread)
        self.installer.finished.connect(self.worker_thread.quit)
        self.installer.progress_update.connect(self._progress_update)
        self.worker_thread.started.connect(self.installer.run)

        self.installing_dialog = fci.loadUi(os.path.join(os.path.dirname(__file__), "progress.ui"))
        self.installing_dialog.setObjectName("AddonManager_InstallingDialog")
        self.installing_dialog.label.setText(
            translate("AddonsInstaller", "Installing '{}'").format(
                self.addon_to_install.display_name
            )
        )

        self.installing_dialog.rejected.connect(self._cancel_addon_installation)
        self.installer.finished.connect(self.installing_dialog.hide)
        self.installing_dialog.show()
        self.worker_thread.start()  # Returns immediately

    def _progress_update(self, bytes_read: int, data_size: int) -> None:
        self.installing_dialog.progressBar.setMaximum(data_size)
        self.installing_dialog.progressBar.setValue(bytes_read)

    def _cancel_addon_installation(self):
        dlg = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.NoIcon,
            translate("AddonsInstaller", "Cancelling"),
            translate("AddonsInstaller", "Cancelling installation of '{}'").format(
                self.addon_to_install.display_name
            ),
            QtWidgets.QMessageBox.NoButton,
            parent=utils.get_main_am_window(),
        )
        dlg.setObjectName("AddonInstaller_CancellingDialog")
        dlg.show()
        if self.worker_thread.isRunning():
            # Interruption can take a second or more, depending on what was being done. Make sure
            # we stay responsive and update the dialog with the text above, etc.
            self.worker_thread.requestInterruption()
            self.worker_thread.quit()
            while self.worker_thread.isRunning():
                self.worker_thread.wait(50)
                QtCore.QCoreApplication.processEvents(QtCore.QEventLoop.AllEvents)
        path = str(os.path.join(self.installer.installation_path, self.addon_to_install.name))
        if os.path.exists(path):
            utils.rmdir(path)
        dlg.hide()
        self.finished.emit()

    def _installation_succeeded(self):
        """Called if the installation was successful."""
        MessageDialog.show_modal(
            MessageDialog.DialogType.INFO,
            "AddonManager_SuccessDialog",
            translate("AddonsInstaller", "Success"),
            translate("AddonsInstaller", "{} was installed successfully").format(
                self.addon_to_install.name
            ),
            QtWidgets.QMessageBox.Ok,
        )
        self.success.emit(self.addon_to_install)
        self.finished.emit()

    def _installation_failed(self, addon, message):
        """Called if the installation failed."""
        error_dialog = QtWidgets.QMessageBox(utils.get_main_am_window())
        error_dialog.setObjectName("AddonManager_ErrorDialog")
        error_dialog.setIcon(QtWidgets.QMessageBox.Critical)
        error_dialog.setWindowTitle(translate("AddonsInstaller", "Installation Failed"))
        error_dialog.setText(
            translate("AddonsInstaller", "Failed to install {}").format(addon.name)
        )
        error_dialog.setStandardButtons(QtWidgets.QMessageBox.Cancel)
        error_dialog.setDetailedText(message)
        error_dialog.exec()
        self.finished.emit()


class MacroInstallerGUI(QtCore.QObject):
    """Install a macro, providing feedback about the process via dialog boxes, and then offer to
    add the macro to a custom toolbar. Should be run on the main GUI thread: this class internally
    launches a QThread for the actual installation process."""

    # Only success should matter to external code: all user interaction is handled via this class
    success: QtCore.Signal = QtCore.Signal(object)

    # Emitted once all work has been completed, regardless of success or failure
    finished: QtCore.Signal = QtCore.Signal()

    def __init__(self, addon):
        """The provided addon object must have an attribute called "macro", and that attribute must
        itself provide a callable "install" method that takes a single string, the path to the
        installation location."""
        super().__init__()
        self.addon_to_install = addon
        self.worker_thread = None
        self.installer = MacroInstaller(self.addon_to_install)
        self.toolbar_adapter = ToolbarAdapter()
        self.macro_dir = fci.DataPaths().macro_dir

    def quit(self):
        if self.worker_thread and hasattr(self.worker_thread, "quit"):
            self.worker_thread.quit()
            self.worker_thread.wait(500)
            if self.worker_thread.isRunning():
                fci.Console.PrintError(
                    "INTERNAL ERROR: Thread did not quit() cleanly, using terminate()\n"
                )
                self.worker_thread.clean_terminate()

    def run(self):
        """Perform the installation, including any necessary user interaction via modal dialog
        boxes. If installation proceeds successfully to completion, emits the success() signal.
        """

        self.worker_thread = QtCore.QThread()
        self.worker_thread.setObjectName("MacroInstallerGUI worker thread")
        self.installer.moveToThread(self.worker_thread)
        self.installer.finished.connect(self.worker_thread.quit)
        self.installer.success.connect(self._base_installation_success)
        self.worker_thread.started.connect(self.installer.run)
        self.worker_thread.start()  # Returns immediately

    def _base_installation_success(self):
        """Callback for a successful basic macro installation."""
        self.success.emit(self.addon_to_install)
        self._ask_to_install_toolbar_button()  # Synchronous set of modals
        self.finished.emit()

    def _ask_to_install_toolbar_button(self) -> None:
        """Presents a dialog to the user asking if they want to install a toolbar button for
        a particular macro and walks through that process if they agree to do so."""
        do_not_show_dialog = fci.Preferences().get("dontShowAddMacroButtonDialog")
        button_exists = self._macro_button_exists()
        if not do_not_show_dialog and not button_exists:
            add_toolbar_button_dialog = fci.loadUi(
                os.path.join(os.path.dirname(__file__), "add_toolbar_button_dialog.ui")
            )
            add_toolbar_button_dialog.setObjectName("AddonManager_AddToolbarButtonDialog")
            add_toolbar_button_dialog.buttonYes.clicked.connect(self._install_toolbar_button)
            add_toolbar_button_dialog.buttonNever.clicked.connect(
                lambda: fci.Preferences().set("dontShowAddMacroButtonDialog", True)
            )
            add_toolbar_button_dialog.exec()

    def _macro_button_exists(self) -> bool:
        """Returns True if a button already exists for this macro, or False if not."""
        command = self.toolbar_adapter.find_custom_command(self.addon_to_install.macro.filename)
        if not command:
            return False
        toolbars = self.toolbar_adapter.get_toolbars()
        for toolbar in toolbars:
            if toolbar.GetString(command, "*") != "*":
                return True
        return False

    def _ask_for_toolbar(self, custom_toolbars):
        """Determine what toolbar to add the icon to. The first time it is called, it prompts the
        user to select or create a toolbar. After that, the prompt is optional and can be configured
        via a preference. Returns the pref group for the new toolbar."""

        # If this is the first time we got to this chunk of code, we are always going to ask.
        first_time = fci.Preferences().get("FirstTimeAskingForToolbar")
        fci.Preferences().set("FirstTimeAskingForToolbar", False)
        ask = fci.Preferences().get("alwaysAskForToolbar")

        if first_time or ask:
            select_toolbar_dialog = fci.loadUi(
                os.path.join(os.path.dirname(__file__), "select_toolbar_dialog.ui")
            )
            select_toolbar_dialog.setObjectName("AddonManager_SelectToolbarDialog")

            select_toolbar_dialog.comboBox.clear()

            for group in custom_toolbars:
                name = self.toolbar_adapter.get_toolbar_name(group)
                if name:
                    select_toolbar_dialog.comboBox.addItem(name)
                else:
                    fci.Console.PrintWarning(
                        f"Custom toolbar {group} does not have a Name element\n"
                    )
            new_menubar_option_text = translate("AddonsInstaller", "Create new toolbar")
            select_toolbar_dialog.comboBox.addItem(new_menubar_option_text)

            result = select_toolbar_dialog.exec()
            if result == QtWidgets.QDialog.Accepted:
                selection = select_toolbar_dialog.comboBox.currentText()
                if select_toolbar_dialog.checkBox.checkState() == QtCore.Qt.Unchecked:
                    fci.Preferences().set("alwaysAskForToolbar", False)
                else:
                    fci.Preferences().set("alwaysAskForToolbar", True)
                if selection == new_menubar_option_text:
                    return self.toolbar_adapter.create_new_custom_toolbar()
                return self.toolbar_adapter.get_toolbar_with_name(selection)
            return None

        # If none of the above code returned...
        custom_toolbar_name = fci.Preferences().get("CustomToolbarName")
        toolbar = self.toolbar_adapter.get_toolbar_with_name(custom_toolbar_name)
        if not toolbar:
            # They told us not to ask, but then the toolbar got deleted... ask anyway!
            fci.Preferences().rem("alwaysAskForToolbar")
            return self._ask_for_toolbar(custom_toolbars)
        return toolbar

    def _install_toolbar_button(self) -> None:
        """If the user has requested a toolbar button be installed, this function is called
        to continue the process and request any additional required information."""
        custom_toolbar_name = fci.Preferences().get("CustomToolbarName")

        # Default to false here: if the variable hasn't been set, we don't assume
        # that we have to ask, because the simplest is to just create a new toolbar
        # and never ask at all.
        ask = fci.Preferences().get("alwaysAskForToolbar")

        # See if there is already a custom toolbar for macros:
        custom_toolbars = self.toolbar_adapter.get_toolbars()
        custom_toolbar = None
        if custom_toolbars:
            # If there are already custom toolbars, see if one of them is the one we used last time
            found_toolbar = False
            for toolbar_pref_group in custom_toolbars:
                name = self.toolbar_adapter.get_toolbar_name(toolbar_pref_group)
                if name == custom_toolbar_name:
                    custom_toolbar = toolbar_pref_group
                    found_toolbar = True
                    break
            if ask or not found_toolbar:
                # We have to ask the user what to do...
                custom_toolbar = self._ask_for_toolbar(custom_toolbars)
                if custom_toolbar:
                    custom_toolbar_name = custom_toolbar.GetString("Name")
                    fci.Preferences().set("CustomToolbarName", custom_toolbar_name)
        else:
            # Create a custom toolbar
            custom_toolbar = self.toolbar_adapter.create_new_custom_toolbar()

        if custom_toolbar:
            self._install_macro_to_toolbar(custom_toolbar)
        else:
            fci.Console.PrintMessage("In the end, no custom toolbar was set, bailing out\n")

    def _install_macro_to_toolbar(self, toolbar) -> None:
        """Adds an icon for the given macro to the given toolbar."""
        menu_text = self.addon_to_install.display_name
        tooltip_text = f"<b>{self.addon_to_install.display_name}</b>"
        if self.addon_to_install.macro.comment:
            tooltip_text += f"<br/><p>{self.addon_to_install.macro.comment}</p>"
            whats_this_text = self.addon_to_install.macro.comment
        else:
            whats_this_text = translate(
                "AddonsInstaller", "A macro installed with the FreeCAD Addon Manager"
            )
        status_tip_text = (
            translate("AddonsInstaller", "Run", "Indicates a macro that can be 'run'")
            + " "
            + self.addon_to_install.display_name
        )
        try:
            if self.addon_to_install.macro.icon:
                _, ext = os.path.splitext(self.addon_to_install.macro.icon)
                extension = ext[1:].lower() if ext else "png"
                if self.addon_to_install.macro.icon_data:
                    macro_name = self.addon_to_install.macro.name
                    icon_file = os.path.normpath(
                        os.path.join(self.macro_dir, f"{macro_name}_icon.{extension}")
                    )
                    with open(icon_file, "wb") as f:
                        f.write(self.addon_to_install.macro.icon_data)
                    pixmap_text = icon_file
                else:
                    fci.Console.PrintMessage(
                        f"No cached icon data for {self.addon_to_install.macro.name}\n"
                    )
                    pixmap_text = None
            elif self.addon_to_install.macro.xpm:
                icon_file = os.path.normpath(
                    os.path.join(self.macro_dir, self.addon_to_install.macro.name + "_icon.xpm")
                )
                with open(icon_file, "w", encoding="utf-8") as f:
                    f.write(self.addon_to_install.macro.xpm)
                pixmap_text = icon_file
            else:
                pixmap_text = None
        except OSError:
            fci.Console.PrintWarning(
                f"Could not create icon for {self.addon_to_install.macro.name} in {self.macro_dir}\n"
            )
            pixmap_text = None

        # Add this command to that toolbar
        self.toolbar_adapter.create_custom_command(
            toolbar,
            self.addon_to_install.macro.filename,
            menu_text,
            tooltip_text,
            whats_this_text,
            status_tip_text,
            pixmap_text,
        )


class AddonDependencyInstallerGUI(QtCore.QObject):

    cancel: QtCore.Signal = QtCore.Signal()
    proceed: QtCore.Signal = QtCore.Signal()

    def __init__(self, addons: List[Addon], deps: MissingDependencies):
        super().__init__()
        self.addons: List[Addon] = addons
        self.deps: MissingDependencies = deps
        self.dependency_worker_thread: QtCore.QThread = None
        self.installer: AddonInstaller = None
        self.dependency_installer: DependencyInstaller = None

        self.dependency_dialog = fci.loadUi(
            os.path.join(os.path.dirname(__file__), "dependency_resolution_dialog.ui")
        )
        self.dependency_dialog.setObjectName("AddonManager_DependencyResolutionDialog")

    def shutdown(self):
        try:
            self._stop_thread(self.dependency_worker_thread)
        except RuntimeError:
            # In some circumstances during shutdown the underlying C++ thread may already be gone:
            # we don't care
            pass

    @staticmethod
    def _stop_thread(thread: QtCore.QThread):
        if thread and hasattr(thread, "quit"):
            if thread.isRunning():
                fci.Console.PrintMessage(
                    "INTERNAL ERROR: a QThread is still running when it should have finished"
                )

                thread.requestInterruption()
                thread.wait(100)
            thread.quit()
            thread.wait(500)
            if thread.isRunning():
                fci.Console.PrintError(
                    "INTERNAL ERROR: Thread did not quit() cleanly, using terminate()\n"
                )
                thread.terminate()

    def run(self):
        """Instructs this class to begin displaying the necessary dialogs to guide a user through
        a dependency installation sequence. All dialogs are modal."""

        self.installer = AddonInstaller(None)  # Only used for the allowlist

        stop_installation = self._check_python_version()
        if stop_installation:
            self.cancel.emit()
            return

        # Required Python
        if self.deps.python_requires:
            # Disallowed packages:
            stop_installation = self._handle_disallowed_python()
            if stop_installation:
                self.cancel.emit()
                return
            # Allowed but not yet installed is handled below

        # Remove any disallowed packages from the optional list
        if self.deps.python_optional:
            self._clean_up_optional()

        # Missing FreeCAD workbenches
        if self.deps.wbs:
            stop_installation = self._report_missing_workbenches()
            if stop_installation:
                self.cancel.emit()
                return

        # If we have any missing dependencies, display a dialog to the user asking if they want to
        # install them.
        if self.deps.external_addons or self.deps.python_requires or self.deps.python_optional:
            # Recoverable: ask the user if they want to install the missing deps, do so, then
            # proceed with the installation
            self._resolve_dependencies_then_continue()
        else:
            self.proceed.emit()

    def _handle_disallowed_python(self) -> bool:
        """Determine if we are missing any required Python packages that are not in the allowed
        packages list. If so, display a message to the user and return True if they want to cancel.
        Otherwise, return False."""

        bad_packages = []
        for dep in self.deps.python_requires:
            if dep.lower() not in self.installer.allowed_packages:
                bad_packages.append(dep)

        for dep in bad_packages:
            self.deps.python_requires.remove(dep)

        if bad_packages:
            # pylint: disable=line-too-long
            if len(self.addons) == 1:
                core_message = translate(
                    "AddonsInstaller",
                    "This addon requires Python packages that are not installed, and cannot be "
                    "installed automatically. To use this addon you must install the following "
                    "Python packages manually:",
                )
            else:
                core_message = translate(
                    "AddonsInstaller",
                    "These addons require Python packages that are not installed, and cannot be "
                    "installed automatically. To use them you must install the following Python "
                    "packages manually:",
                )
            message = f"<p>{core_message}</p><ul>"
            if len(bad_packages) < 15:
                for dep in bad_packages:
                    message += f"<li>{dep}</li>"
            else:
                message += "<li>(" + translate("AddonsInstaller", "Too many to list") + ")</li>"
            message += "</ul>"
            message += "To ignore this error and install anyway, press OK."

            r = MessageDialog.show_modal(
                MessageDialog.DialogType.ERROR,
                "AddonManager_RequirementFailedDialog",
                translate("AddonsInstaller", "Requirement Cannot be Installed"),
                message,
                QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
            )
            if r == QtWidgets.QMessageBox.Ok:
                # Force the installation to proceed
                return False
            return True
        return False

    def _report_missing_workbenches(self) -> bool:
        """If there are missing workbenches, display a dialog informing the user. Returns True to
        stop the installation, or False to proceed."""
        addon_name = ""
        if len(self.addons) == 1:
            addon_name = self.addons[0].name
        if len(self.deps.wbs) == 1:
            name = self.deps.wbs[0]
            if addon_name:
                message = translate(
                    "AddonsInstaller",
                    "Addon '{}' requires '{}', which is not available in your copy of FreeCAD.",
                ).format(addon_name, name)
            else:
                message = translate(
                    "AddonsInstaller",
                    "These addons require '{}', which is not available in your copy of FreeCAD.",
                ).format(name)
        else:
            if addon_name:
                message = (
                    "<p>"
                    + translate(
                        "AddonsInstaller",
                        "Addon '{}' requires the following workbenches, which are not available in "
                        "your copy of FreeCAD:",
                    ).format(addon_name)
                    + "</p><ul>"
                )
            else:
                message = (
                    "<p>"
                    + translate(
                        "AddonsInstaller",
                        "These addons require the following workbenches, which are not available in"
                        " your copy of FreeCAD:",
                    )
                    + "</p><ul>"
                )
            for wb in self.deps.wbs:
                message += "<li>" + wb + "</li>"
            message += "</ul>"
            message += translate("AddonsInstaller", "Press OK to install anyway.")

        r = MessageDialog.show_modal(
            MessageDialog.DialogType.ERROR,
            "AddonManager_MissingRequirementDialog",
            translate("AddonsInstaller", "Missing Requirement"),
            message,
            QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel,
        )
        return r == QtWidgets.QMessageBox.Cancel

    def _resolve_dependencies_then_continue(self) -> None:
        """Ask the user how they want to handle dependencies, do that, then install."""

        for addon in self.deps.external_addons:
            self.dependency_dialog.listWidgetAddons.addItem(addon.display_name)
        for mod in self.deps.python_requires:
            self.dependency_dialog.listWidgetPythonRequired.addItem(mod)
        for mod in self.deps.python_optional:
            item = QtWidgets.QListWidgetItem(mod)
            item.setFlags(item.flags() | QtCore.Qt.ItemIsUserCheckable)
            item.setCheckState(QtCore.Qt.Unchecked)
            self.dependency_dialog.listWidgetPythonOptional.addItem(item)

        self.dependency_dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Yes).clicked.connect(
            self._dependency_dialog_yes_clicked
        )
        self.dependency_dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Ignore).clicked.connect(
            self.proceed.emit
        )
        self.dependency_dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).clicked.connect(
            self.cancel.emit
        )
        self.dependency_dialog.buttonBox.button(QtWidgets.QDialogButtonBox.Cancel).setDefault(True)
        self.dependency_dialog.exec()

    def _check_python_version(self) -> bool:
        """Make sure we have a compatible Python version. Returns True to stop the installation
        or False to continue."""

        # For now only look at the minor version, since major is always Python 3
        python_version = Version(
            from_list=[sys.version_info.major, sys.version_info.minor, sys.version_info.micro]
        )
        if python_version < self.deps.python_min_version:
            MessageDialog.show_modal(
                MessageDialog.DialogType.ERROR,
                "AddonManager_IncompatiblePythonVersionDialog",
                translate("AddonsInstaller", "Incompatible Python version"),
                translate(
                    "AddonsInstaller",
                    "This addon (or one of its dependencies) requires Python {}, and your system"
                    " is running {}. Installation cancelled.",
                ).format(str(self.deps.python_min_version), str(python_version)),
                QtWidgets.QMessageBox.Cancel,
            )

            return True
        return False

    def _clean_up_optional(self):
        good_packages = []
        for dep in self.deps.python_optional:
            if dep in self.installer.allowed_packages:
                good_packages.append(dep)
            else:
                fci.Console.PrintWarning(
                    translate(
                        "AddonsInstaller",
                        "Optional dependency on {} ignored because it is not in the allow-list",
                    ).format(dep)
                    + "\n"
                )
        self.deps.python_optional = good_packages

    def _dependency_dialog_yes_clicked(self) -> None:
        addons = self.deps.external_addons
        python_requires = self.deps.python_requires
        python_optional = []
        for row in range(self.dependency_dialog.listWidgetPythonOptional.count()):
            item = self.dependency_dialog.listWidgetPythonOptional.item(row)
            if item.checkState() == QtCore.Qt.Checked:
                python_optional.append(item.text())

        if len(addons) + len(python_requires) + len(python_optional) == 0:
            # No need to run the installer, the user didn't select anything to install
            self.proceed.emit()
        else:
            self._run_dependency_installer(addons, python_requires, python_optional)

    def _run_dependency_installer(self, addons, python_requires, python_optional):
        """Run the dependency installer (in a separate thread) for the given dependencies"""
        self.dependency_installer = DependencyInstaller(addons, python_requires, python_optional)
        self.dependency_installer.no_python_exe.connect(self._report_no_python_exe)
        self.dependency_installer.no_pip.connect(self._report_no_pip)
        self.dependency_installer.failure.connect(self._report_dependency_failure)
        self.dependency_installer.finished.connect(self._dependencies_finished)

        self.dependency_worker_thread = QtCore.QThread(self)
        self.dependency_worker_thread.setObjectName("Dependency Installer Thread")
        self.dependency_installer.moveToThread(self.dependency_worker_thread)
        self.dependency_worker_thread.started.connect(self.dependency_installer.run)
        self.dependency_installer.finished.connect(self.dependency_worker_thread.quit)

        self.dependency_installation_dialog = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.Information,
            translate("AddonsInstaller", "Installing Dependencies", "Window title"),
            translate("AddonsInstaller", "Installing dependencies…", "Window text"),
            QtWidgets.QMessageBox.Cancel,
            parent=utils.get_main_am_window(),
        )
        self.dependency_installation_dialog.setObjectName(
            "AddonManager_InstallingDependenciesDialog"
        )
        self.dependency_installation_dialog.rejected.connect(self._cancel_dependency_installation)
        self.dependency_installation_dialog.show()
        self.dependency_worker_thread.start()

    def _report_no_python_exe(self) -> None:
        """Callback for the dependency installer failing to locate a Python executable."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        result = MessageDialog.show_modal(
            MessageDialog.DialogType.ERROR,
            "AddonManager_CannotExecutePythonDialog",
            translate("AddonsInstaller", "Cannot execute Python"),
            translate(
                "AddonsInstaller",
                "Failed to automatically locate your Python executable, or the path is set "
                "incorrectly. Please check the Addon Manager preferences setting for the path to "
                "Python.",
            )
            + "\n\n"
            + translate(
                "AddonsInstaller",
                "Dependencies could not be installed. Continue with installation anyway?",
            ),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        )
        if result == QtWidgets.QMessageBox.Yes:
            self.proceed.emit()
        else:
            self.cancel.emit()

    def _report_no_pip(self, command: str) -> None:
        """Callback for the dependency installer failing to access pip."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        result = MessageDialog.show_modal(
            MessageDialog.DialogType.ERROR,
            "AddonManager_CannotExecutePipDialog",
            translate("AddonsInstaller", "Cannot execute pip"),
            translate(
                "AddonsInstaller",
                "Failed to execute pip, which may be missing from your Python installation. Please "
                "ensure your system has pip installed and try again. The failed command was:",
            )
            + f" \n\n{command}\n\n"
            + translate(
                "AddonsInstaller",
                "Continue with addon installation anyway?",
            ),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        )
        if result == QtWidgets.QMessageBox.Yes:
            self.proceed.emit()
        else:
            self.cancel.emit()

    def _report_dependency_failure(self, short_message: str, details: str) -> None:
        """Callback for dependency installation failure."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        if self.dependency_installer and hasattr(self.dependency_installer, "finished"):
            self.dependency_installer.finished.disconnect(self._report_dependency_success)
        fci.Console.PrintError(details + "\n")
        result = MessageDialog.show_modal(
            MessageDialog.DialogType.ERROR,
            "AddonManager_PackageInstallationFailedDialog",
            translate("AddonsInstaller", "Package installation failed"),
            short_message
            + "\n\n"
            + translate("AddonsInstaller", "See Report View for detailed failure log.")
            + "\n\n"
            + translate(
                "AddonsInstaller",
                "Continue with installation anyway?",
            ),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
        )
        if result == QtWidgets.QMessageBox.Yes:
            self.proceed.emit()
        else:
            self.cancel.emit()

    def _report_dependency_success(self):
        """Callback for dependency installation success."""
        if self.dependency_installation_dialog is not None:
            self.dependency_installation_dialog.hide()
        self.proceed.emit()

    def _dependencies_finished(self, success: bool):
        if success:
            self._report_dependency_success()
        self.dependency_worker_thread.quit()

    def _cancel_dependency_installation(self) -> None:
        """Cancel was clicked in the dialog. NOTE: Does no cleanup, the state after cancellation is
        unknown. In most cases pip can recover from whatever we've done to it."""
        self.dependency_worker_thread.blockSignals(True)
        self.dependency_installer.blockSignals(True)
        self.dependency_worker_thread.requestInterruption()
        self.dependency_installation_dialog.hide()
        self.dependency_worker_thread.wait()
        self.cancel.emit()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
from typing import Optional
from PySide import QtGui, QtCore
from ..models.base import ToolBitShape


def _png2qpixmap(data, icon_size):
    pixmap = QtGui.QPixmap()
    pixmap.loadFromData(data, "PNG")
    # Scale the pixmap if the requested size is different
    if pixmap.size() != icon_size:
        pixmap = pixmap.scaled(
            icon_size,
            QtCore.Qt.KeepAspectRatio,
            QtCore.Qt.SmoothTransformation,
        )
    return pixmap


class ShapeWidget(QtGui.QWidget):
    def __init__(self, shape: ToolBitShape, icon_size: Optional[QtCore.QSize] = None, parent=None):
        super(ShapeWidget, self).__init__(parent)
        self.layout = QtGui.QVBoxLayout(self)
        self.layout.setAlignment(QtCore.Qt.AlignHCenter)

        self.shape = shape
        self.icon_size = icon_size or QtCore.QSize(263, 372)  # A4 aspect ratio
        self.icon_widget = QtGui.QLabel()
        self.layout.addWidget(self.icon_widget)

        self._update_icon()

    def _update_icon(self):
        ratio = self.devicePixelRatioF()
        size = self.icon_size * ratio
        icon = self.shape.get_icon()
        if icon:
            pixmap = icon.get_qpixmap(size)
            self.icon_widget.setPixmap(pixmap)
            return

        thumbnail = self.shape.get_thumbnail()
        if thumbnail:
            pixmap = _png2qpixmap(thumbnail, size)
            self.icon_widget.setPixmap(pixmap)
            return

        self.icon_widget.clear()  # Clear pixmap if no icon



# ***************************************************************************
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com>   *
# *   Copyright (c) 2021 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import sys
import FreeCAD
from FreeCAD import Rotation
from FreeCAD import Vector

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Electrostatics Capacitance Two Balls",
        "meshtype": "solid",
        "meshelement": "Tet4",
        "constraints": ["electrostatic potential"],
        "solvers": ["elmer"],
        "material": "fluid",
        "equations": ["electrostatic"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.equation_electrostatics_capacitance_two_balls import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=41488&start=90#p412047

Electrostatics equation in FreeCAD FEM-Elmer

"""
    )


def setup(doc=None, solvertype="elmer"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects
    small_sphere1 = doc.addObject("Part::Sphere", "Small_Sphere1")
    small_sphere1.Placement = FreeCAD.Placement(Vector(-1000, 0, 0), Rotation(Vector(0, 0, 1), 0))
    small_sphere1.Radius = "500 mm"

    small_sphere2 = doc.addObject("Part::Sphere", "Small_Sphere2")
    small_sphere2.Placement = FreeCAD.Placement(Vector(1000, 0, 0), Rotation(Vector(0, 0, 1), 0))
    small_sphere2.Radius = "500 mm"

    fusion = doc.addObject("Part::MultiFuse", "Fusion")
    fusion.Shapes = [small_sphere1, small_sphere2]

    large_sphere = doc.addObject("Part::Sphere", "Large_Sphere")
    large_sphere.Radius = "5000 mm"

    geom_obj = doc.addObject("Part::Cut", "Cut")
    geom_obj.Base = large_sphere
    geom_obj.Tool = fusion
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Transparency = 75
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")
    if FreeCAD.GuiUp:
        import FemGui

        FemGui.setActiveAnalysis(analysis)

    # solver
    if solvertype == "elmer":
        solver_obj = ObjectsFem.makeSolverElmer(doc, "SolverElmer")
        eq_electrostatic = ObjectsFem.makeEquationElectrostatic(doc, solver_obj)
        eq_electrostatic.CalculateCapacitanceMatrix = True
        eq_electrostatic.CalculateElectricEnergy = True
        eq_electrostatic.CalculateElectricField = True
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialFluid(doc, "Air")
    mat = material_obj.Material
    mat["Name"] = "Air"
    mat["Density"] = "1.204 kg/m^3"
    mat["ThermalConductivity"] = "0.02587 W/m/K"
    mat["ThermalExpansionCoefficient"] = "3.43e-3 1/K"
    mat["SpecificHeat"] = "1.01 kJ/kg/K"
    mat["ElectricalConductivity"] = "1e-12 S/m"
    mat["RelativePermeability"] = "1.0"
    mat["RelativePermittivity"] = "1.00059"
    material_obj.Material = mat
    material_obj.References = [(geom_obj, "Solid1")]
    analysis.addObject(material_obj)

    # constraint potential 1st
    name_pot1 = "ElectrostaticPotential1"
    con_elect_pot1 = ObjectsFem.makeConstraintElectrostaticPotential(doc, name_pot1)
    con_elect_pot1.References = [(geom_obj, "Face1")]
    con_elect_pot1.ElectricInfinity = True
    con_elect_pot1.PotentialEnabled = False
    analysis.addObject(con_elect_pot1)

    # constraint potential 2nd
    name_pot2 = "ElectrostaticPotential2"
    con_elect_pot2 = ObjectsFem.makeConstraintElectrostaticPotential(doc, name_pot2)
    con_elect_pot2.References = [(geom_obj, "Face2")]
    con_elect_pot2.CapacitanceBody = 1
    con_elect_pot2.CapacitanceBodyEnabled = True
    con_elect_pot2.PotentialEnabled = False
    analysis.addObject(con_elect_pot2)

    # constraint potential 3rd
    name_pot3 = "ElectrostaticPotential3"
    con_elect_pot3 = ObjectsFem.makeConstraintElectrostaticPotential(doc, name_pot3)
    con_elect_pot3.References = [(geom_obj, "Face3")]
    con_elect_pot3.CapacitanceBody = 2
    con_elect_pot3.CapacitanceBodyEnabled = True
    con_elect_pot3.PotentialEnabled = False
    analysis.addObject(con_elect_pot3)

    # mesh
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False
    femmesh_obj.CharacteristicLengthMax = "600 mm"
    femmesh_obj.ViewObject.Visibility = False

    # mesh_region
    mesh_region = ObjectsFem.makeMeshRegion(doc, femmesh_obj, name="MeshRegion")
    mesh_region.CharacteristicLength = "250 mm"
    mesh_region.References = [(geom_obj, "Face2"), (geom_obj, "Face3")]
    mesh_region.ViewObject.Visibility = False

    # generate the mesh
    success = generate_mesh.mesh_from_mesher(femmesh_obj, "gmsh")
    if not success:
        # try to create from existing rough mesh
        from .meshes.mesh_capacitance_two_balls_tetra10 import (
            create_nodes,
            create_elements,
        )

        fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
        femmesh_obj.FemMesh = fem_mesh

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2025 Furgo                                              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import FreeCAD as App
import Arch
from bimtests import TestArchBase


class TestArchBuildingPart(TestArchBase.TestArchBase):

    def testMakeFloorEmpty(self):
        floor = Arch.makeFloor()
        self.assertIsNotNone(floor, "Failed to create an empty floor")

    def testMakeFloorWithObjects(self):
        obj = App.ActiveDocument.addObject("Part::Box", "Box")
        floor = Arch.makeFloor([obj])
        self.assertIn(obj, floor.Group, "Object not added to the floor")

    def testFloorProperties(self):
        floor = Arch.makeFloor()
        self.assertEqual(floor.Label, "Level", "Default label is incorrect")

    def testFloor(self):
        App.Console.PrintLog("Checking Arch Floor...\n")
        structure = Arch.makeStructure(length=2, width=3, height=5)
        floor = Arch.makeFloor([structure])
        self.assertTrue(floor, "Arch Floor failed")

    def testBuilding(self):
        App.Console.PrintLog("Checking Arch Building...\n")
        structure = Arch.makeStructure(length=2, width=3, height=5)
        floor = Arch.makeFloor([structure])
        building = Arch.makeBuilding([floor])
        self.assertTrue(building, "Arch Building failed")

    def testSite(self):
        App.Console.PrintLog("Checking Arch Site...\n")
        structure = Arch.makeStructure(length=2, width=3, height=5)
        floor = Arch.makeFloor([structure])
        building = Arch.makeBuilding([floor])
        site = Arch.makeSite([building])
        self.assertTrue(site, "Arch Site failed")

    def test_makeBuildingPart(self):
        """Test the makeBuildingPart function."""
        operation = "Testing makeBuildingPart function"
        self.printTestMessage(operation)

        part = Arch.makeBuildingPart(name="TestBuildingPart")
        self.assertIsNotNone(part, "makeBuildingPart failed to create a building part.")
        self.assertEqual(part.Label, "TestBuildingPart", "Building part label is incorrect.")

    def test_makeFloor(self):
        """Test the makeFloor function."""
        operation = "Testing makeFloor function"
        self.printTestMessage(operation)

        floor = Arch.makeFloor(name="TestFloor")
        self.assertIsNotNone(floor, "makeFloor failed to create a floor object.")
        self.assertEqual(floor.Label, "TestFloor", "Floor label is incorrect.")

    def test_makeBuilding(self):
        """Test the makeBuilding function."""
        operation = "Testing makeBuilding function"
        self.printTestMessage(operation)

        building = Arch.makeBuilding(name="TestBuilding")
        self.assertIsNotNone(building, "makeBuilding failed to create a building object.")
        self.assertEqual(building.Label, "TestBuilding", "Building label is incorrect.")

    def test_convertFloors(self):
        """Test the convertFloors function."""
        operation = "Testing convertFloors..."
        self.printTestMessage(operation)

        # Create a mock floor object
        floor = Arch.makeFloor()
        Arch.convertFloors(floor)
        self.assertEqual(
            floor.IfcType, "Building Storey", "convertFloors failed to set IfcType correctly"
        )

    def test_make2DDrawing(self):
        """Test the make2DDrawing function."""
        operation = "Testing make2DDrawing..."
        self.printTestMessage(operation)

        obj = Arch.make2DDrawing()
        self.assertIsNotNone(obj, "make2DDrawing failed to create an object")
        self.assertEqual(obj.Label, "Drawing", "Incorrect default label for 2D Drawing")



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "Open files FEM App unit tests"
__author__ = "Bernd Hahnebach"
__url__ = "https://www.freecad.org"

import platform
import tempfile
import unittest
from os.path import join
from os import remove

import FreeCAD

from . import support_utils as testtools
from .support_utils import fcc_print


"""
FIXME TODO HACK
Important note!
Delete build directory (at least in fem the objects and vpobjects directories)
if not migrate will not be used because the old modules might still be in the
build directory, thus the test will fail
rm -rf Mod/Fem/
FIXME TODO HACK
"""


"""
# TODO: separate unit test:
# std document name of object ==  obj type
ATM:
for elmer equation obj: name != proxy type
material solid and material fluid obj: name != proxy type

# in addition for FeaturePythons
# std document name of object ==  the class name (for for femsolver obj)
# all femsolver objects class name is Proxy
"""


class TestObjectOpen(unittest.TestCase):
    fcc_print("import TestObjectOpen")

    # ********************************************************************************************
    def setUp(self):
        # setUp is executed before every test

        # new document
        self.document = FreeCAD.newDocument(self.__class__.__name__)

        self.test_file_dir = join(testtools.get_fem_test_home_dir(), "open")

    # ********************************************************************************************
    def tearDown(self):
        # tearDown is executed after every test
        FreeCAD.closeDocument(self.document.Name)

    # ********************************************************************************************
    def test_00print(self):
        # since method name starts with 00 this will be run first
        # this test just prints a line with stars

        fcc_print(
            "\n{0}\n{1} run FEM TestObjectOpen tests {2}\n{0}".format(100 * "*", 10 * "*", 60 * "*")
        )

    # ********************************************************************************************
    def test_femobjects_open_head(self):
        fcc_print("load master head document objects")

        # get a document with all FEM objects
        from .test_object import create_all_fem_objects_doc

        self.document = create_all_fem_objects_doc(self.document)

        # save and load the document
        file_path = join(tempfile.gettempdir(), "all_objects_head.FCStd")
        self.document.saveAs(file_path)
        FreeCAD.closeDocument(self.document.Name)
        self.document = FreeCAD.open(file_path)

        # C++ objects
        self.compare_cpp_objs(self.document)
        # FeaturePythons objects
        self.compare_feature_pythons_class_app(self.document)

        # standard name changed
        from femsolver.elmer.equations.flux import Proxy

        remove(file_path)
        self.assertEqual(Proxy, self.document.Flux.Proxy.__class__)

    # ********************************************************************************************
    def test_femobjects_open_de9b3fb438(self):
        # migration modules fail on s390x (big endian) and trigger OOMs,
        # https://bugs.debian.org/984952 and
        # https://bugs.launchpad.net/ubuntu/+source/freecad/+bug/1918474.
        if platform.machine() == "s390x":
            return

        # the number in method name is the FreeCAD commit the document was created with
        # https://github.com/FreeCAD/FreeCAD/commit/de9b3fb438
        # the document was created by running the object create unit test
        # FreeCAD --run-test "femtest.app.test_object.TestObjectCreate.test_femobjects_make"
        fcc_print("load old document objects")
        FreeCAD.closeDocument(self.document.Name)  # close the empty document from setUp first
        self.document = FreeCAD.open(join(self.test_file_dir, "all_objects_de9b3fb438.FCStd"))

        # C++ objects
        self.compare_cpp_objs(self.document)
        # FeaturePythons objects
        self.compare_feature_pythons_class_app(self.document)

        # standard name changed
        from femsolver.elmer.equations.flux import Proxy

        self.assertEqual(Proxy, self.document.Fluxsolver.Proxy.__class__)

    # ********************************************************************************************
    def compare_cpp_objs(self, doc):
        from femtools.femutils import type_of_obj

        self.assertEqual("Fem::FemAnalysis", type_of_obj(doc.Analysis))
        # TODO other C++ objects and view provider
        # Is just checking the type sufficient?
        # If there is a type there is at least a object with correct type ;-)

    # ********************************************************************************************
    def compare_feature_pythons_class_app(self, doc):
        import ObjectsFem
        from femtools.femutils import type_of_obj

        # see comments at file end, the code was created by some python code
        """
        # see code lines after comment block for the smarter version
        # but this makes it easy to understand what is happening
        self.assertEqual(
            "<class 'femobjects.constraint_bodyheatsource.ConstraintBodyHeatSource'>",
            str(doc.ConstraintBodyHeatSource.Proxy.__class__)
        )
        """
        from femobjects.constraint_bodyheatsource import ConstraintBodyHeatSource

        self.assertEqual(ConstraintBodyHeatSource, doc.ConstraintBodyHeatSource.Proxy.__class__)

        self.assertEqual(
            "Fem::ConstraintCurrentDensity",
            type_of_obj(ObjectsFem.makeConstraintCurrentDensity(doc)),
        )

        from femobjects.constraint_electrostaticpotential import ConstraintElectrostaticPotential

        self.assertEqual(
            ConstraintElectrostaticPotential, doc.ConstraintElectrostaticPotential.Proxy.__class__
        )

        from femobjects.constraint_flowvelocity import ConstraintFlowVelocity

        self.assertEqual(ConstraintFlowVelocity, doc.ConstraintFlowVelocity.Proxy.__class__)

        from femobjects.constraint_initialflowvelocity import ConstraintInitialFlowVelocity

        self.assertEqual(
            ConstraintInitialFlowVelocity, doc.ConstraintInitialFlowVelocity.Proxy.__class__
        )

        self.assertEqual(
            "Fem::ConstraintMagnetization", type_of_obj(ObjectsFem.makeConstraintMagnetization(doc))
        )

        from femobjects.constraint_selfweight import ConstraintSelfWeight

        self.assertEqual(ConstraintSelfWeight, doc.ConstraintSelfWeight.Proxy.__class__)

        from femobjects.constraint_tie import ConstraintTie

        self.assertEqual(ConstraintTie, doc.ConstraintTie.Proxy.__class__)

        from femobjects.element_fluid1D import ElementFluid1D

        self.assertEqual(ElementFluid1D, doc.ElementFluid1D.Proxy.__class__)

        from femobjects.element_geometry1D import ElementGeometry1D

        self.assertEqual(ElementGeometry1D, doc.ElementGeometry1D.Proxy.__class__)

        from femobjects.element_geometry2D import ElementGeometry2D

        self.assertEqual(ElementGeometry2D, doc.ElementGeometry2D.Proxy.__class__)

        from femobjects.element_rotation1D import ElementRotation1D

        self.assertEqual(ElementRotation1D, doc.ElementRotation1D.Proxy.__class__)

        from femobjects.material_common import MaterialCommon

        self.assertEqual(MaterialCommon, doc.MaterialFluid.Proxy.__class__)

        from femobjects.material_common import MaterialCommon

        self.assertEqual(MaterialCommon, doc.MaterialSolid.Proxy.__class__)

        from femobjects.material_mechanicalnonlinear import MaterialMechanicalNonlinear

        self.assertEqual(
            MaterialMechanicalNonlinear, doc.MaterialMechanicalNonlinear.Proxy.__class__
        )

        from femobjects.material_reinforced import MaterialReinforced

        self.assertEqual(MaterialReinforced, doc.MaterialReinforced.Proxy.__class__)

        from femobjects.mesh_gmsh import MeshGmsh

        self.assertEqual(MeshGmsh, doc.MeshGmsh.Proxy.__class__)

        from femobjects.mesh_boundarylayer import MeshBoundaryLayer

        self.assertEqual(MeshBoundaryLayer, doc.MeshBoundaryLayer.Proxy.__class__)

        from femobjects.mesh_group import MeshGroup

        self.assertEqual(MeshGroup, doc.MeshGroup.Proxy.__class__)

        from femobjects.mesh_region import MeshRegion

        self.assertEqual(MeshRegion, doc.MeshRegion.Proxy.__class__)

        from femobjects.mesh_result import MeshResult

        self.assertEqual(MeshResult, doc.MeshResult.Proxy.__class__)

        from femobjects.result_mechanical import ResultMechanical

        self.assertEqual(ResultMechanical, doc.ResultMechanical.Proxy.__class__)

        from femobjects.solver_ccxtools import SolverCcxTools

        self.assertEqual(SolverCcxTools, doc.SolverCcxTools.Proxy.__class__)

        from femobjects.solver_elmer import SolverElmer

        self.assertEqual(SolverElmer, doc.SolverElmer.Proxy.__class__)

        from femsolver.z88.solver import Proxy

        self.assertEqual(Proxy, doc.SolverZ88.Proxy.__class__)

        from femsolver.elmer.equations.elasticity import Proxy

        self.assertEqual(Proxy, doc.Elasticity.Proxy.__class__)

        from femsolver.elmer.equations.electrostatic import Proxy

        self.assertEqual(Proxy, doc.Electrostatic.Proxy.__class__)

        from femsolver.elmer.equations.flow import Proxy

        self.assertEqual(Proxy, doc.Flow.Proxy.__class__)

        from femsolver.elmer.equations.heat import Proxy

        self.assertEqual(Proxy, doc.Heat.Proxy.__class__)

        self.assertEqual(
            "Fem::EquationElmerMagnetodynamic2D",
            type_of_obj(ObjectsFem.makeEquationMagnetodynamic2D(doc)),
        )

        self.assertEqual(
            "Fem::EquationElmerMagnetodynamic",
            type_of_obj(ObjectsFem.makeEquationMagnetodynamic(doc)),
        )


"""
# code was generated by the following code from a document with all objects
# run test_object.test_femobjects_make how to create such document
# in tmp in FEM_unittests will be the file with all objects
# the doc.Name of some objects needs to be edited after run
# because obj. name != proxy type
# elmer equation objects need to be edited after
# material solid and material fluid

# objects
from femtools.femutils import type_of_obj
for o in App.ActiveDocument.Objects:
    if hasattr(o, "Proxy"):
        module_with_class = str(o.Proxy.__class__).lstrip("<class '").rstrip("'>")
        class_of_o = module_with_class.split(".")[-1]
        o_name_from_type = type_of_obj(o).lstrip('Fem::')
        module_to_load = module_with_class.rstrip(class_of_o).rstrip(".")
        print("        from {} import {}".format(module_to_load, class_of_o))
        print("        self.assertEqual(")
        print("            {},".format(class_of_o))
        print("            doc.{}.Proxy.__class__".format(o_name_from_type))
        print("        )")
        print("")

for o in App.ActiveDocument.Objects:
    if hasattr(o, "Proxy"):
        o.Proxy.__class__

"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2025 Samuel Abels <knipknap@gmail.com>                  *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
import uuid
import pathlib
from typing import Mapping, Union, Optional, List, Dict, cast
import Path
from ...assets import Asset, AssetUri
from ...toolbit import ToolBit


class Library(Asset):
    asset_type: str = "toolbitlibrary"
    API_VERSION = 1

    def __init__(self, label, id=None):
        self.id = id if id is not None else str(uuid.uuid4())
        self._label = label
        self._bits: List[ToolBit] = []
        self._bit_nos: Dict[int, ToolBit] = {}
        self._bit_urls: Dict[AssetUri, ToolBit] = {}

    @property
    def label(self) -> str:
        return self._label

    def get_id(self) -> str:
        """Returns the unique identifier for the Library instance."""
        return self.id

    @classmethod
    def resolve_name(cls, identifier: Union[str, AssetUri, pathlib.Path]) -> AssetUri:
        """
        Resolves various forms of library identifiers to a canonical AssetUri string.
        Handles direct AssetUri objects, URI strings, asset IDs, or legacy filenames.
        Returns the canonical URI string or None if resolution fails.
        """
        if isinstance(identifier, AssetUri):
            return identifier

        if isinstance(identifier, str) and AssetUri.is_uri(identifier):
            return AssetUri(identifier)

        if isinstance(identifier, pathlib.Path):  # Handle direct Path objects (legacy filenames)
            identifier = identifier.stem  # Use the filename stem as potential ID

        if not isinstance(identifier, str):
            raise ValueError("Failed to resolve {identifier} to a Uri")

        return AssetUri.build(asset_type=Library.asset_type, asset_id=identifier)

    def to_dict(self) -> dict:
        """Returns a dictionary representation of the Library in the specified format."""
        tools_list = []
        for tool_no, tool in self._bit_nos.items():
            tools_list.append(
                {"nr": tool_no, "path": f"{tool.get_id()}.fctb"}  # Tool ID with .fctb extension
            )
        return {"label": self.label, "tools": tools_list, "version": self.API_VERSION}

    @classmethod
    def from_dict(
        cls,
        data_dict: dict,
        id: str,
        dependencies: Optional[Mapping[AssetUri, Asset]],
    ) -> "Library":
        """
        Creates a Library instance from a dictionary and resolved dependencies.
        If dependencies is None, it's a shallow load, and tools are not populated.
        """
        library = cls(data_dict.get("label", id or "Unnamed Library"), id=id)

        if dependencies is None:
            Path.Log.debug(
                f"Library.from_dict: Shallow load for library '{library.label}' (id: {id}). Tools not populated."
            )
            return library  # Only process tools if dependencies were resolved

        tools_list = data_dict.get("tools", [])
        for tool_data in tools_list:
            tool_no = tool_data["nr"]
            tool_id = pathlib.Path(tool_data["path"]).stem  # Extract tool ID
            tool_uri = AssetUri(f"toolbit://{tool_id}")
            bit = cast(ToolBit, dependencies.get(tool_uri))
            if bit:
                library.add_bit(bit, bit_no=tool_no)
            else:
                raise ValueError(f"Tool with id {tool_id} not found in dependencies")
        return library

    def __str__(self):
        return '{} "{}"'.format(self.id, self.label)

    def __eq__(self, other):
        return self.id == other.id

    def __iter__(self):
        return self._bits.__iter__()

    def get_next_bit_no(self):
        bit_nolist = sorted(self._bit_nos, reverse=True)
        return bit_nolist[0] + 1 if bit_nolist else 1

    def get_bit_no_from_bit(self, bit: ToolBit) -> Optional[int]:
        for bit_no, thebit in self._bit_nos.items():
            if bit == thebit:
                return bit_no
        return None

    def get_tool_by_uri(self, uri: AssetUri) -> Optional[ToolBit]:
        for tool in self._bit_nos.values():
            if tool.get_uri() == uri:
                return tool
        return None

    def assign_new_bit_no(self, bit: ToolBit, bit_no: Optional[int] = None) -> int:
        if bit not in self._bits:
            raise ValueError(f"given bit {bit} not in library; cannot assign tool number")

        # If no specific bit_no was requested, assign a new one.
        if bit_no is None:
            bit_no = self.get_next_bit_no()
        elif self._bit_nos.get(bit_no) == bit:
            return bit_no

        # Otherwise, add the bit. Since the requested bit_no may already
        # be in use, we need to account for that. In this case, we will
        # add the removed bit into a new bit_no.
        old_bit = self._bit_nos.pop(bit_no, None)
        old_bit_no = self.get_bit_no_from_bit(bit)
        if old_bit_no:
            del self._bit_nos[old_bit_no]
        self._bit_nos[bit_no] = bit
        if old_bit:
            self.assign_new_bit_no(old_bit)
        return bit_no

    def add_bit(self, bit: ToolBit, bit_no: Optional[int] = None) -> int:
        if bit not in self._bits:
            self._bits.append(bit)
        return self.assign_new_bit_no(bit, bit_no)

    def get_bits(self) -> List[ToolBit]:
        return self._bits

    def has_bit(self, bit: ToolBit) -> bool:
        for t in self._bits:
            if bit.id == t.id:
                return True
        return False

    def remove_bit(self, bit: ToolBit):
        self._bits = [t for t in self._bits if t.id != bit.id]
        self._bit_nos = {k: v for (k, v) in self._bit_nos.items() if v.id != bit.id}

    def remove_bit_by_uri(self, uri: AssetUri | str):
        if isinstance(uri, str):
            uri = AssetUri(uri)
        self._bits = [t for t in self._bits if t.get_uri() != uri]
        self._bit_nos = {k: v for (k, v) in self._bit_nos.items() if v.get_uri() != uri}

    def dump(self, summarize: bool = False):
        title = 'Library "{}" ({}) (instance {})'.format(self.label, self.id, id(self))
        print("-" * len(title))
        print(title)
        print("-" * len(title))
        for bit in self._bits:
            print(f"- {bit.label} ({bit.get_id()})")
        print()



#! python
# (c) 2010 Werner Mayer LGPL

__author__ = "Werner Mayer <wmayer[at]users.sourceforge.net>"

# Formulas:
# M2 = P + b*r2 + t*u
# S1 = (r2*M1 + r1*M2)/(r1+r2)
# S2 = M2-b*r2

import math


# 3d vector class
class Vector:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z

    def add(self, vec):
        return Vector(self.x + vec.x, self.y + vec.y, self.z + vec.z)

    def sub(self, vec):
        return Vector(self.x - vec.x, self.y - vec.y, self.z - vec.z)

    def dot(self, vec):
        return self.x * vec.x + self.y * vec.y + self.z * vec.z

    def mult(self, s):
        return Vector(self.x * s, self.y * s, self.z * s)

    def cross(self, vec):
        return Vector(
            self.y * vec.z - self.z * vec.y,
            self.z * vec.x - self.x * vec.z,
            self.x * vec.y - self.y * vec.x,
        )

    def length(self):
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    def norm(self):
        l = self.length()
        if l > 0:
            self.x /= l
            self.y /= l
            self.z /= l

    def __repr__(self):
        return "(%f,%f,%f)" % (self.x, self.y, self.z)


# A signum function
def sgn(val):
    if val > 0:
        return 1
    elif val < 0:
        return -1
    else:
        return 0


# M1  ... is the center of the arc
# P   ... is the end point of the arc and start point of the line
# Q   ..  is a second point on the line
# N   ... is the normal of the plane where the arc and the line lie on, usually N=(0,0,1)
# r2  ... the fillet radius
# ccw ... counter-clockwise means which part of the arc is given. ccw must be either True or False


def makeFilletArc(M1, P, Q, N, r2, ccw):
    u = Q.sub(P)
    v = P.sub(M1)
    if ccw:
        b = u.cross(N)
    else:
        b = N.cross(u)
    b.norm()

    uu = u.dot(u)
    uv = u.dot(v)
    r1 = v.length()

    # distinguish between internal and external fillets
    r2 *= sgn(uv)

    cc = 2.0 * r2 * (b.dot(v) - r1)
    dd = uv * uv - uu * cc
    if dd < 0:
        raise RuntimeError("Unable to calculate intersection points")
    t1 = (-uv + math.sqrt(dd)) / uu
    t2 = (-uv - math.sqrt(dd)) / uu

    if abs(t1) < abs(t2):
        t = t1
    else:
        t = t2

    br2 = b.mult(r2)
    print(br2)
    ut = u.mult(t)
    print(ut)
    M2 = P.add(ut).add(br2)
    S1 = M1.mult(r2 / (r1 + r2)).add(M2.mult(r1 / (r1 + r2)))
    S2 = M2.sub(br2)

    return (S1, S2, M2)


def test():
    from FreeCAD import Base
    import Part

    P1 = Base.Vector(1, -5, 0)
    P2 = Base.Vector(-5, 2, 0)
    P3 = Base.Vector(1, 5, 0)
    # Q = Base.Vector(5, 10, 0)
    # Q = Base.Vector(5, 11, 0)
    Q = Base.Vector(5, 0, 0)
    r2 = 3.0
    axis = Base.Vector(0, 0, 1)
    ccw = False

    arc = Part.ArcOfCircle(P1, P2, P3)
    C = arc.Center
    Part.show(Part.makeLine(P3, Q))
    Part.show(arc.toShape())

    (S1, S2, M2) = makeArc(
        Vector(C.x, C.y, C.z),
        Vector(P3.x, P3.y, P3.z),
        Vector(Q.x, Q.y, Q.z),
        Vector(axis.x, axis.y, axis.z),
        r2,
        ccw,
    )
    circle = Part.Circle(Base.Vector(M2.x, M2.y, M2.z), Base.Vector(0, 0, 1), math.fabs(r2))
    Part.show(circle.toShape())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2022 - 2025 Larry Woestman <LarryWoestman2@gmail.com>   *
# *   Copyright (c) 2024 Ondsel <development@ondsel.com>                    *
# *   Copyright (c) 2024 Carl Slater <CandLWorkshopLLC@gmail.com>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

from typing import Any, Dict

from Path.Post.scripts.refactored_linuxcnc_post import Refactored_Linuxcnc

import Path
import FreeCAD

translate = FreeCAD.Qt.translate

DEBUG = False
if DEBUG:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

#
# Define some types that are used throughout this file.
#
Values = Dict[str, Any]


class Refactored_Masso_G3(Refactored_Linuxcnc):
    """The Refactored Masso G3 post processor class."""

    def __init__(
        self,
        job,
        tooltip=translate("CAM", "Refactored Masso G3 post processor"),
        tooltipargs=[""],
        units="Metric",
    ) -> None:
        super().__init__(
            job=job,
            tooltip=tooltip,
            tooltipargs=tooltipargs,
            units=units,
        )
        Path.Log.debug("Refactored Masso G3 post processor initialized.")

    def init_values(self, values: Values) -> None:
        """Initialize values that are used throughout the postprocessor."""
        #
        super().init_values(values)
        #
        # Set any values here that need to override the default values set
        # in the parent routine.
        #
        #
        # Used in the argparser code as the "name" of the postprocessor program.
        #
        values["MACHINE_NAME"] = "Masso G3"
        values["POSTPROCESSOR_FILE_NAME"] = __name__
        #
        # setting TOOL_BEFORE_CHANGE to True will output T# M6 before each tool change
        # rather than M6 T#.
        #
        values["TOOL_BEFORE_CHANGE"] = True

    @property
    def tooltip(self):
        tooltip: str = """
        This is a postprocessor file for the CAM workbench.
        It is used to take a pseudo-gcode fragment from a CAM object
        and output 'real' GCode suitable for a Masso G3 3 axis mill.
        """
        return tooltip



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the object code for the BezCurve object."""
## @package bezcurve
# \ingroup draftobjects
# \brief Provides the object code for the BezCurve object.

## \addtogroup draftobjects
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
from draftutils import gui_utils
from draftutils import params
from draftobjects.base import DraftObject


class BezCurve(DraftObject):
    """The BezCurve object"""

    def __init__(self, obj):
        super().__init__(obj, "BezCurve")

        _tip = QT_TRANSLATE_NOOP("App::Property", "The points of the Bezier curve")
        obj.addProperty("App::PropertyVectorList", "Points", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The degree of the Bezier function")
        obj.addProperty("App::PropertyInteger", "Degree", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Continuity")
        obj.addProperty("App::PropertyIntegerList", "Continuity", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "If the Bezier curve should be closed or not")
        obj.addProperty("App::PropertyBool", "Closed", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "Create a face if this curve is closed")
        obj.addProperty("App::PropertyBool", "MakeFace", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The length of this object")
        obj.addProperty("App::PropertyLength", "Length", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The area of this object")
        obj.addProperty("App::PropertyArea", "Area", "Draft", _tip, locked=True)

        obj.MakeFace = params.get_param("MakeFaceMode")
        obj.Closed = False
        obj.Degree = 3
        obj.Continuity = []
        # obj.setEditorMode("Degree", 2)
        obj.setEditorMode("Continuity", 1)

    def onDocumentRestored(self, obj):
        super().onDocumentRestored(obj)
        gui_utils.restore_view_object(
            obj, vp_module="view_bezcurve", vp_class="ViewProviderBezCurve"
        )

    def execute(self, fp):
        if self.props_changed_placement_only():
            fp.positionBySupport()
            self.props_changed_clear()
            return

        self.createGeometry(fp)
        fp.positionBySupport()
        self.props_changed_clear()

    def _segpoleslst(self, fp):
        """Split the points into segments."""
        if not fp.Closed and len(fp.Points) >= 2:  # allow lower degree segment
            poles = fp.Points[1:]
        elif fp.Closed and len(fp.Points) >= fp.Degree:  # drawable
            # poles=fp.Points[1:(fp.Degree*(len(fp.Points)//fp.Degree))]+fp.Points[0:1]
            poles = fp.Points[1:] + fp.Points[0:1]
        else:
            poles = []
        return [poles[x : x + fp.Degree] for x in range(0, len(poles), (fp.Degree or 1))]

    def resetcontinuity(self, fp):
        fp.Continuity = [0] * (len(self._segpoleslst(fp)) - 1 + 1 * fp.Closed)
        # nump= len(fp.Points)-1+fp.Closed*1
        # numsegments = (nump // fp.Degree) + 1 * (nump % fp.Degree > 0) -1
        # fp.Continuity = [0]*numsegments

    def onChanged(self, fp, prop):
        self.props_changed_store(prop)

        if prop == "Closed":
            # if remove the last entry when curve gets opened
            oldlen = len(fp.Continuity)
            newlen = len(self._segpoleslst(fp)) - 1 + 1 * fp.Closed
            if oldlen > newlen:
                fp.Continuity = fp.Continuity[:newlen]
            if oldlen < newlen:
                fp.Continuity = fp.Continuity + [0] * (newlen - oldlen)

        if (
            hasattr(fp, "Closed")
            and fp.Closed
            and prop in ["Points", "Degree", "Closed"]
            and len(fp.Points) % fp.Degree
        ):
            # the curve editing tools can't handle extra points
            fp.Points = fp.Points[: (fp.Degree * (len(fp.Points) // fp.Degree))]
            # for closed curves

        if prop in ["Degree"] and fp.Degree >= 1:
            self.resetcontinuity(fp)

        if prop in ["Points", "Degree", "Continuity", "Closed"]:
            self.createGeometry(fp)

    def createGeometry(self, fp):
        import Part

        plm = fp.Placement
        if fp.Points:
            startpoint = fp.Points[0]
            edges = []
            for segpoles in self._segpoleslst(fp):
                #               if len(segpoles) == fp.Degree # would skip additional poles
                c = Part.BezierCurve()  # last segment may have lower degree
                c.increase(len(segpoles))
                c.setPoles([startpoint] + segpoles)
                edges.append(Part.Edge(c))
                startpoint = segpoles[-1]
            w = Part.Wire(edges)
            if fp.Closed and w.isClosed():
                try:
                    if hasattr(fp, "MakeFace"):
                        if fp.MakeFace:
                            w = Part.Face(w)
                    else:
                        w = Part.Face(w)
                except Part.OCCError:
                    pass
            fp.Shape = w
            if hasattr(fp, "Area") and hasattr(w, "Area"):
                fp.Area = w.Area
            if hasattr(fp, "Length") and hasattr(w, "Length"):
                fp.Length = w.Length
        fp.Placement = plm

    @classmethod
    def symmetricpoles(cls, knot, p1, p2):
        """Make two poles symmetric respective to the knot."""
        p1h = App.Vector(p1)
        p2h = App.Vector(p2)
        p1h.multiply(0.5)
        p2h.multiply(0.5)
        return (knot + p1h - p2h, knot + p2h - p1h)

    @classmethod
    def tangentpoles(cls, knot, p1, p2, allowsameside=False):
        """Make two poles have the same tangent at knot."""
        p12n = p2.sub(p1)
        p12n.normalize()
        p1k = knot - p1
        p2k = knot - p2
        p1k_ = App.Vector(p12n)
        kon12 = p1k * p12n
        if allowsameside or not (kon12 < 0 or p2k * p12n > 0):  # instead of moving
            p1k_.multiply(kon12)
            pk_k = knot - p1 - p1k_
            return (p1 + pk_k, p2 + pk_k)
        else:
            return cls.symmetricpoles(knot, p1, p2)

    @staticmethod
    def modifysymmetricpole(knot, p1):
        """calculate the coordinates of the opposite pole
        of a symmetric knot"""
        return knot + knot - p1

    @staticmethod
    def modifytangentpole(knot, p1, oldp2):
        """calculate the coordinates of the opposite pole
        of a tangent knot"""
        pn = knot - p1
        pn.normalize()
        pn.multiply((knot - oldp2).Length)
        return pn + knot


# Alias for compatibility with v0.18 and earlier
_BezCurve = BezCurve

## @}



# ***************************************************************************
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com    *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Rotation
from FreeCAD import Vector

import Part

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Constraint Contact Solid Solid",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["fixed", "pressure", "contact"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.constraint_contact_solid_solid import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?f=18&t=20276
constraint contact for solid to solid mesh

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects
    # bottom box
    bottom_box_obj = doc.addObject("Part::Box", "BottomBox")
    bottom_box_obj.Length = 100
    bottom_box_obj.Width = 25
    bottom_box_obj.Height = 500
    bottom_box_obj.Placement = FreeCAD.Placement(
        Vector(186, 0, -247),
        Rotation(0, 0, 0),
        Vector(0, 0, 0),
    )
    doc.recompute()

    # top half cylinder, https://forum.freecad.org/viewtopic.php?f=18&t=43001#p366111
    top_halfcyl_obj = doc.addObject("Part::Cylinder", "TopHalfCylinder")
    top_halfcyl_obj.Radius = 30
    top_halfcyl_obj.Height = 500
    top_halfcyl_obj.Angle = 180
    top_halfcyl_sh = Part.getShape(top_halfcyl_obj, "", needSubElement=False, refine=True)
    top_halfcyl_obj.Shape = top_halfcyl_sh
    top_halfcyl_obj.Placement = FreeCAD.Placement(
        Vector(0, -42, 0),
        Rotation(0, 90, 0),
        Vector(0, 0, 0),
    )
    doc.recompute()

    # all geom fusion
    geom_obj = doc.addObject("Part::MultiFuse", "AllGeomFusion")
    geom_obj.Shapes = [bottom_box_obj, top_halfcyl_obj]
    geom_obj.Refine = True
    if FreeCAD.GuiUp:
        bottom_box_obj.ViewObject.hide()
        top_halfcyl_obj.ViewObject.hide()
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
        solver_obj.SplitInputWriter = False
        """
        # solver parameter from fandaL, but they are not needed (see forum topic)
        solver_obj.IterationsControlParameterTimeUse = True
        solver_obj.IterationsControlParameterCutb = '0.25,0.5,0.75,0.85,,,1.5,'
        solver_obj.IterationsControlParameterIter = '4,8,9,200,10,400,,200,,'
        solver_obj.IterationsUserDefinedTimeStepLength = True
        solver_obj.TimeInitialStep = 0.1
        solver_obj.TimeEnd = 1.0
        solver_obj.IterationsUserDefinedIncrementations = True  # parameter DIRECT
        """
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "MechanicalMaterial")
    mat = material_obj.Material
    mat["Name"] = "Steel-Generic"
    mat["YoungsModulus"] = "200000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [
        (geom_obj, "Face5"),
        (geom_obj, "Face6"),
        (geom_obj, "Face8"),
        (geom_obj, "Face9"),
    ]
    analysis.addObject(con_fixed)

    # constraint pressure
    con_pressure = ObjectsFem.makeConstraintPressure(doc, "ConstraintPressure")
    con_pressure.References = [(geom_obj, "Face10")]
    con_pressure.Pressure = "100.0 MPa"
    con_pressure.Reversed = False
    analysis.addObject(con_pressure)

    # constraint contact
    con_contact = ObjectsFem.makeConstraintContact(doc, "ConstraintContact")
    con_contact.References = [
        (geom_obj, "Face7"),  # first seems slave face, TODO proof in writer code!
        (geom_obj, "Face3"),  # second seems master face, TODO proof in writer code!
    ]
    con_contact.Friction = False
    con_contact.Slope = "1000000.0 GPa/m"
    analysis.addObject(con_contact)

    # mesh
    from .meshes.mesh_contact_box_halfcylinder_tetra10 import (
        create_nodes,
        create_elements,
    )

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False

    doc.recompute()
    return doc
